#include "h2o.h"
#include <math.h>

/* external variables definition */
float VFP,VGP,RHOF,RHOG,HFP,HGP,SFP,SGP,EFP,EGP;
float VOL,TOTM,TOTH,TOTS,TOTE,X,VOID,RHO,SPECVOL;
float SPECH,SPECS,SPECE,PS,TS;
float SRFDP,SRFDT,SRGDP,SRGDT,SHFDP,SHFDT;
float SHGDP,SHGDT,G1P,G2P,G1T,G2T,F1,F2,F3,F4,F5;
float STDP,SPDT,SRDP,SRDT,SHDP,SHDT;
float SSFDP,SSFDT,SSGDP,SSGDT,SSDP,SSDT;
float CPFP,CPGP,VISCFP,VISCGP,BETAFP,BETAGP,AKFP,AKGP;
float CP,VISC,BETA,AK,SPEEDFP,SPEEDGP,SPEED,CVFP,CVGP,CV;
float TC,TCFP,TCGP,SIGMA;
char *ERRH2O;
#define MIN(a,b)  ((a<=b) ? (a) : (b))
#define MAX(a,b)  ((a>=b) ? (a) : (b))
#define ABS(a)    ((a>=0) ? (a) : -(a))

int PROP(float *pP,float *pT,int ISAT,int ITYPE,int NMAX,
     float ERRTOL,char *ERRMSG,float ADJ)
/************************************************************************
*   VERSION 0.0                                                         *
*     THIS PROGRAM GENERATES H2O PROPERTIES                             *
*************************************************************************
*     INPUT FLAGS:                                                      *     
*                                                                       *
*                                            ITYPE                      *
*                                                                       *
*                        0       1       2       3        4             *
*       ISAT                                                            *
*       ----------------------------------------------------------------* 
*         0       based on (P,T) independent of ITYPE                   *
*         1       (P,??)  (P,RHO)  (P,h)   (P,x)   (P,void)             *
*         2       (??,T)  (T,RHO)  (T,h)   (T,x)   (T,void)             *
*         3       based on (RHO,h) independent of ITYPE                 *
*         4       based on (RHO,P) independent of ITYPE                 *
*         5       based on (RHO,T) independent of ITYPE                 * 
*         6       based on (h,P) independent of ITYPE                   *  
*         7       based on (h,T) independent of ITYPE                   *   
*                                                                       *  
*         ?? MEANS DON'T CARE                                           *
*                                                                       *
*       OUTPUT FLAG:                                                    * 
*                            IFLAG - 0 = SUBCOOLED                      *
*                                    1 = SATURATED                      *
*                                    2 = SUPERHEATED                    *
*************************************************************************
*       INITIALIZATION                                                  *
************************************************************************/
{

/* function prototypes */
    float PSAT(float T);
    float TSAT(float P);
    void CLIPP(float *pP);
    void CLIPT(float *pT);
    void CLIPH(float *pH);
    void CLIPR(float *pR);
    void SUB(float P,float T);
    void SAT(float P, unsigned ITYPE);
    void SUP(float P,float T);
    void GSUB(float P,float T,float *GP1,float *GP2,float *GT1,float *GT2);
    void GSUP(float P,float T,float *GP1,float *GP2,float *GT1,float *GT2);
    float HF(float P);
    float HG(float P);
    float HFSUB(float P,float T);
    float HGSUP(float P,float T);
    float DF(float P);
    float DG(float P);
    float DFSUB(float P,float T);
    float DGSUP(float P,float T);
    float DRFDTP(float P,float T);
    float DRGDTP(float P,float T);
    float DRFDPT(float P,float T);
    float DRGDPT(float P,float T);
    float DHFDTP(float P,float T);
    float DHGDTP(float P,float T);
    float DHFDPT(float P,float T);
    float DHGDPT(float P,float T);
/* variables declarations */
    float P,T;
    float SPECH0,RHO1,SPECH1,DR,DH,GG1P,GG2P,GG1T,GG2T,DP,DT;
    float ERROR1,ERROR2,ERROR;
    unsigned IFLAG,N,NOGO;
        P = *pP;
        T = *pT;
/************************************************************************
*       THE SIMPLEST CASE IS WHEN ISAT=0: P & T ARE GIVEN, THE FLUID IS   *
*       ASSUMED TO NOT BE SATURATED AND THE PROPERTIES ARE CALCULATED     *
*       DIRECTLY. THERE IS A CHANCE THAT THE FLUID WILL TURN OUT TO BE    *
*       SATURATED. IF THIS IS SO, THEN ISAT IS SET TO 1 AND PROPERTY      *
*       EVALUATION IS DEFERRED TILL LATER (THE CASE OF ISAT=1, WITH       *
*       ITYPE AND VAR USED AS PREVIOUSLY DEFINED (DEFAULT=0).             *
*************************************************************************
*       FIND THE SATURATION POINTS CORRESPONDING TO THE INPUT T & P       *
************************************************************************/
      PS=PSAT(T);
      TS=TSAT(P);
      CLIPP(&P);
      CLIPT(&T);
      CLIPH(&SPECH);
      CLIPR(&RHO);
/************************************************************************
*       P & T GIVEN                                                     *
************************************************************************/

      if (ISAT == 0)
      {         
          if (P > PS)                      /* SUBCOOLED */
          {       
                    IFLAG=0;
                    SUB(P,T);
          }       
          else if (P < PS)        /* SUPERHEATED */
          {
                    IFLAG=2;
                    SUP(P,T);
          }
          else  /*       RARE CASE OF SATURATED FLUID FOR GIVEN P & T */
          {
                    ISAT=1;
                    ITYPE=0;
          }
       }
          
/************************************************************************
*       SATURATED BASED ON P
************************************************************************/
          else if (ISAT == 1)
          {
                IFLAG=1;
                T=TSAT(P);
                SAT(P,ITYPE);
          }
          
/************************************************************************
*       SATURATED BASED ON T
************************************************************************/
      else if (ISAT == 2)
      {
                IFLAG=1;
                P=PSAT(T);
                SAT(P,ITYPE);
          }
/************************************************************************
*       RHO & h GIVEN
************************************************************************/
      else if (ISAT == 3)
      {
                N=0;
                NOGO=0;
Label20:        N=N+1;
                HGP=HG(P);
                HFP=HF(P);
                if(SPECH < HFP)
                {
                        IFLAG=0;
                        RHO1=DFSUB(P,T);
                        SPECH1=HFSUB(P,T);
                        DR=RHO-RHO1;
                        DH=SPECH-SPECH1;
                        GSUB(P,T,&GG1P,&GG2P,&GG1T,&GG2T);
                        DP=GG1P*DR+GG2P*DH;
                        DT=GG1T*DR+GG2T*DH;
                        P=P+DP*ADJ;
                        T=T+DT*ADJ;
                }
                else if(SPECH > HGP)
                {
                        IFLAG=2;
                        RHO1=DGSUP(P,T);
                        SPECH1=HGSUP(P,T);
                        DR=RHO-RHO1;
                        DH=SPECH-SPECH1;
                        GSUP(P,T,&GG1P,&GG2P,&GG1T,&GG2T);
                        DP=GG1P*DR+GG2P*DH;
                        DT=GG1T*DR+GG2T*DH;
                        P=P+DP*ADJ;
                        T=T+DT*ADJ;
                }
		else      /* saturation */
		{
			IFLAG=1;
			DR=0.;
                        SPECH0=SPECH;
                        ITYPE=1;
                        SAT(P,ITYPE);
                        SPECH1=SPECH;
                        SPECH=SPECH0;
                        DH=SPECH-SPECH1;
                        DP=DH/SHDP;
                        DT=0.;
                        P=P+DP*ADJ;
                        T=TSAT(P);
                }
          
                CLIPP(&P);
		CLIPT(&T);
		ERROR1=MAX(ABS(DR/RHO),ABS(DH/SPECH));
		ERROR2=MAX(ABS(DP/P),ABS(DT/T));
		ERROR=MAX(ERROR1,ERROR2);
                
		if(ERROR <= ERRTOL) NOGO=1;
                
                if(N >= NMAX)
                {
			NOGO=1;
                        ERRMSG="MAXIMUM NUMBER OF ITERATIONS USED";
                }
                
                if (NOGO == 0) goto Label20;
                if (IFLAG == 0)   SUB(P,T);
                if (IFLAG == 1)   SAT(P,ITYPE);
                if (IFLAG == 2)   SUP(P,T);
        }
/************************************************************************
*       RHO & P GIVEN
************************************************************************/
                
                else if (ISAT == 4)   
                {
                N=0;
		NOGO=0;
                T=TSAT(P);                       /* new */
                if(RHO > DF(P))
                {
Label30:                N=N+1;
                        IFLAG=0;
                        RHO1=DFSUB(P,T);
                        DR=RHO-RHO1;
                        DT=DR/DRFDTP(P,T);
                        ERROR=MAX(ABS(DR/RHO),ABS(DT/T));
                        if(ERROR <= ERRTOL) NOGO=1;
                        if(N >= NMAX)
                        {
                           NOGO=1;
                           ERRMSG=" MAXIMUM NUMBER OF ITERATIONS USED";
                        }
                        T=T+DT*ADJ;
                        CLIPT(&T);
                        if (NOGO == 0) goto Label30;
                        SUB(P,T);
                }
                else if(RHO < DG(P))
                {
Label40:                N=N+1;
                        IFLAG=2;
                        RHO1=DGSUP(P,T);
                        DR=RHO-RHO1;
                        DT=DR/DRGDTP(P,T);
                        ERROR=MAX(ABS(DR/RHO),ABS(DT/T));
                        if(ERROR <= ERRTOL) NOGO=1;
                        if(N >= NMAX)
                        {
                           NOGO=1;
                           ERRMSG=" MAXIMUM NUMBER OF ITERATIONS USED";
                        }
                        T=T+DT*ADJ;
                        CLIPT(&T);
                        if (NOGO == 0) goto Label40;
                          SUP(P,T);
                }
                else
                {
                        T=TSAT(P);
                        IFLAG=1;
                        ITYPE=1;
                        SAT(P,ITYPE);
                }
        }
        
/************************************************************************
*       RHO & T GIVEN
************************************************************************/
        
        else if (ISAT == 5)
        {
                PS=PSAT(T);
                N=0;
                NOGO=0;
                P=PS;                            /* new */
                if(RHO > DF(PS)) 
                {
Label50:                N=N+1;
                        IFLAG=0;
                        RHO1=DFSUB(P,T);
                        DR=RHO-RHO1;
                        DP=DR/DRFDPT(P,T);
                        ERROR=MAX(ABS(DR/RHO),ABS(DP/P));
                        if(ERROR <= ERRTOL) NOGO=1;
                        if(N >= NMAX)
                        {
                            NOGO=1;
                            ERRMSG=" MAXIMUM NUMBER OF ITERATIONS USED";
                        }
                        P=P+DP*ADJ;
                        CLIPP(&P);
                        if (NOGO == 0) goto Label50;
                        SUB(P,T);
                }
                else if(RHO < DG(PS))
                {
Label60:                N=N+1;
                        IFLAG=2;
                        RHO1=DGSUP(P,T);
                        DR=RHO-RHO1;
                        DP=DR/DRGDPT(P,T);
                        ERROR=MAX(ABS(DR/RHO),ABS(DP/P));
                        if(ERROR <= ERRTOL) NOGO=1;
                        if(N >= NMAX)
                        {
                            NOGO=1;
                            ERRMSG=" MAXIMUM NUMBER OF ITERATIONS USED";
                        }
                        P=P+DP*ADJ;
                        CLIPP(&P);
                        if (NOGO == 0) goto Label60;
                        SUP(P,T);
                }
                else
                {
                        IFLAG=1;
                        P=PS;
                        ITYPE=1;
                        SAT(P,ITYPE);
                }
        }
        
/************************************************************************
*       h & P GIVEN
************************************************************************/
        
        else if (ISAT == 6)
        {
                N=0;
                NOGO=0;
                T=TSAT(P);                             /* new */
                if(SPECH < HF(P))
                {
Label70:                N=N+1;
                        IFLAG=0;
                        SPECH1=HFSUB(P,T);
                        DH=SPECH-SPECH1;
                        DT=DH/DHFDTP(P,T);
                        ERROR=MAX(ABS(DH/SPECH),ABS(DT/T));
                        if(ERROR <= ERRTOL) NOGO=1;
                        if(N >= NMAX)
                        {
                           NOGO=1;
                           ERRMSG=" MAXIMUM NUMBER OF ITERATIONS USED";
                        }
                        T=T+DT*ADJ;
                        CLIPT(&T);
                        if (NOGO == 0) goto Label70;
                        SUB(P,T);
                }
                else if(SPECH > HG(P))
                {
Label80:                N=N+1;
                        IFLAG=2;
                        SPECH1=HGSUP(P,T);
                        DH=SPECH-SPECH1;
                        DT=DH/DHGDTP(P,T);
                        ERROR=MAX(ABS(DH/SPECH),ABS(DT/T));
                        if(ERROR <= ERRTOL) NOGO=1;
                        if(N >= NMAX)
                        {
                                NOGO=1;
                                ERRMSG=" MAXIMUM NUMBER OF ITERATIONS USED";
                        }
                        T=T+DT*ADJ;
                        CLIPT(&T);
                        if (NOGO == 0) goto Label80;
                        SUP(P,T);
                }
                else
                {
                        T=TSAT(P);
                        IFLAG=1;
                        ITYPE=2;
                        SAT(P,ITYPE);
                }
        }
        
/************************************************************************
*       h & T GIVEN
************************************************************************/
        else if (ISAT == 7)
        {
                PS=PSAT(T);
                N=0;
                NOGO=0;
                P=PS;                                 /* new */
                if(SPECH < HF(PS))
                {
Label90:                N=N+1;
                        IFLAG=0;
                        SPECH1=HFSUB(P,T);
                        DH=SPECH-SPECH1;
                        DP=DH/DHFDPT(P,T);
                        ERROR=MAX(ABS(DH/SPECH),ABS(DP/P));
                        if(ERROR <= ERRTOL) NOGO=1;
                        if(N >= NMAX)
                        {
                            NOGO=1;
                            ERRMSG=" MAXIMUM NUMBER OF ITERATIONS USED";
                        }
                        P=P+DP*ADJ;
                        CLIPP(&P);
                        if (NOGO == 0) goto Label90;
                        SUB(P,T);
                }
                else if(SPECH > HG(PS))
                {
Label99:                N=N+1;
                        IFLAG=2;
                        SPECH1=HGSUP(P,T);
                        DH=SPECH-SPECH1;
                        DP=DH/DHGDPT(P,T);
                        ERROR=MAX(ABS(DH/SPECH),ABS(DP/P));
                        if(ERROR <= ERRTOL) NOGO=1;
                        if(N >= NMAX)
                        {
                            NOGO=1;
                            ERRMSG=" MAXIMUM NUMBER OF ITERATIONS USED";
                        }
                        P=P+DP*ADJ;
                        CLIPP(&P);
                        if (NOGO == 0) goto Label99;
                        SUP(P,T);
                }
                else
                {
                        IFLAG=1;
                        P=PS;
                        ITYPE=2;
                        SAT(P,ITYPE);
                }
        }
                *pP=P;
                *pT=T;
        NMAX=N;
        return(IFLAG);
} /* end of function PROP */

/************************************************************************
* SATURATED PROPERTY ROUTINE
************************************************************************/
void SAT(float P,unsigned ITYPE)
{
    /* function prototypes */
    
    float VF(float P);
    float VG(float P);
    float HF(float P);
    float HG(float P);
    float SF(float P);
    float SG(float P);
    float CPF(float P);
    float CPG(float P);
    float DRFDP(float P);
    float DVFDP(float P);
    float DHFDP(float P);
    float DSFDP(float P);
    float DRGDP(float P);
    float DVGDP(float P);
    float DHGDP(float P);
    float DSGDP(float P);
    float DTDP(float P);
    float DPDT(float P);
    float VISCF(float P);
    float VISCG(float P);
    float TCF(float P);
    float TCG(float P);
    float SIGMAF(float T);
    /* variable declarations */
    float SVFDP,SVGDP,SVDP,DENOM,TK,RATIO;
/************************************************************************
*       FLAGS:                                                                                             
*
*               ITYPE - 0=ONLY PRESSURE AND TEMPERATURE KNOWN                                      *
*               1=DENSITY KNOWN                                                                            
*               2=SPECIFIC ENTHALPY KNOWN                                                                  *
*               3=QUALITY KNOWN                                                                            
*               4=VOID KNOWN
*
************************************************************************/
        
        VFP=VF(P);
        VGP=VG(P);
        HFP=HF(P);
        HGP=HG(P);
        SFP=SF(P);
        SGP=SG(P);
        EFP=HF(P)-P*VFP*1000;
        EGP=HG(P)-P*VGP*1000;
        RHOG=1./VGP;
        RHOF=1./VFP;
/************************************************************************
*  P OR T KNOWN                                                                                            
************************************************************************/
/************************************************************************
* IF DENSITY, ENTHALPY, QUALITY OR VOID NOT SPECIFIED, THEN CAN'T         * 
* COMPLETE CALCULATIONS----assume single phase liquid and proceed         *
************************************************************************/
        if (ITYPE == 0)
        {
                X=0.0;
                VOID=0.0;
                RHO=RHOF;
                SPECVOL=VFP;
                SPECH=HFP;
                SPECE=EFP;
                SPECS=SFP;
                ERRH2O="PHASE QUANTITIES CALCULATED BASED ON X=0";
        }
        
/************************************************************************
* RHO KNOWN                                                                                                
************************************************************************/
          
        else if (ITYPE == 1)
        {
                VOID=(RHOF-RHO)/(RHOF-RHOG);
                SPECVOL=1.0/RHO;
                X=(VFP-SPECVOL)/(VFP-VGP);
                SPECH=HFP + X * (HGP-HFP);
                SPECE=EFP + X * (EGP-EFP);
                SPECS=SFP + X * (SGP-SFP);
        }
               
/************************************************************************
*  h KNOWN                                                                                                 
************************************************************************/
        
        else if (ITYPE == 2)
        {
                X=(SPECH-HFP)/(HGP-HFP);
                SPECVOL=VFP + X * (VGP-VFP);
                RHO=1/SPECVOL;
                VOID=(RHOF-RHO)/(RHOF-RHOG);
                SPECE=EFP + X * (EGP-EFP);
                SPECS=SFP + X * (SGP-SFP);
        }
        
/************************************************************************
*  X KNOWN
************************************************************************/
        
        else if (ITYPE == 3)
        {
                SPECVOL=VFP + X * (VGP-VFP);
                RHO=1/SPECVOL;
                VOID=(RHOF-RHO)/(RHOF-RHOG);
                SPECH=HFP + X * (HGP-HFP);
                SPECE=EFP + X * (EGP-EFP);
                SPECS=SFP + X * (SGP-SFP);
        }
        
/************************************************************************
* VOID KNOWN
************************************************************************/
        
        else if (ITYPE == 4)
        {
                RHO=RHOF + VOID * (RHOG-RHOF);
                SPECVOL=1/RHO;
                X=(VFP-SPECVOL)/(VFP-VGP);
                SPECH=HFP + X * (HGP-HFP);
                SPECE=EFP + X * (EGP-EFP);
                SPECS=SFP + X * (SGP-SFP);
        }
        
/* EXTENSIVE PROPERTIES */
        TOTM=RHO*VOL;
        TOTH=SPECH*TOTM;
        TOTE=SPECE*TOTM;
        TOTS=SPECS*TOTM;
/* PROPERTIES ALONG THE SATURATION LINE */
        TS    = TSAT(P);
        STDP  = DTDP(P);
        SPDT  = DPDT(TS);
        SHFDP = DHFDP(P);
        SVFDP = DVFDP(P);
        SRFDP = -RHOF*RHOF * DVFDP(P);
        SSFDP = DSFDP(P);
        SHGDP = DHGDP(P);
        SVGDP = DVGDP(P);
        SRGDP = -RHOG*RHOG * DVGDP(P);
        SSGDP = DSGDP(P);
        SHFDT = DHFDP(P) * SPDT;
        SRFDT = SRFDP    * SPDT;
        SHGDT = DHGDP(P) * SPDT;
        SRGDT = SRGDP    * SPDT;
        SSFDT = DSFDP(P) * SPDT;
        SSGDT = DSGDP(P) * SPDT;
	SHDP  = SHFDP * (1.0-X) + SHGDP * X - (HGP-HFP)/(VGP-VFP) *
               ( X * (SVGDP - SVFDP) + SVFDP);
	SSDP  = SSFDP * (1.0-X) + SSGDP * X - (SGP-SFP)/(VGP-VFP) *
               ( X * (SVGDP - SVFDP) + SVFDP);
	SVDP  = SVFDP * (1.0-X) + SVGDP * X - (VGP-VFP)/(HGP-HFP) *
               ( X * (SHGDP - SHFDP) + SHFDP);
        SRDP  = -RHO*RHO*SVDP;
        SRDT  = SRDP * SPDT;
        SHDT  = SHDP * SPDT;
        SSDT  = SSDP * SPDT;
/* F FUNCTIONS */
        F1= HGP*VFP -HFP*VGP;
        F2= VGP-VFP;
        F3= HFP -HGP;
        F4= SHGDP*( VGP-VFP) - SVGDP*( HGP-HFP);
        F5= SHFDP*( VGP-VFP) - SVFDP*( HGP-HFP);
/* G FUNCTIONS */
	DENOM= X * F4 + (1.0-X) * F5;
        G1P= -F3 * pow(SPECVOL,2) / DENOM;
        G2P= F2 / DENOM;
        G1T= G1P * STDP;
        G2T= G2P * STDP;
/* BETA */
        BETAFP = -DRFDTP(P,TS)/RHOF;  /* beta=(1/v)*(dv/dT)|p in 1/K */
        BETAGP = -DRGDTP(P,TS)/RHOG;
        BETA   =  (1.-X)*BETAFP + X*BETAGP;       /* new !!! */
/* Ak */
        AKFP = DRFDPT(P,TS)/RHOF;      /* alfa =-(1/v)*(dv/dp)|T in 1/MPa */
        AKGP = DRGDPT(P,TS)/RHOG;
        AK   = (1.-X)*AKFP + X*AKGP;              /* new !!! */
/* Cp */
        CPFP = CPF(P);
        CPGP = CPG(P);
/* Mass weighted Cp */
	CP = CPFP*(1.0-X) + CPGP*X;
/* Cv */
	TK = TS + 273.16;
	CVFP = CPFP - BETAFP*BETAFP * TK * VFP / AKFP *1.E3;
	CVGP = CPGP - BETAGP*BETAGP * TK * VGP / AKGP *1.E3;
/* Mass weighted Cv */
	CV = CVFP*(1.0-X) + CVGP*X;
/* DYNAMIC VISCOSITY */
        VISCFP = VISCF(P);
        VISCGP = VISCG(P);
/*   DYNAMIC VISCOSITY OF THE MIXTURE  ??? - new !!! */
	 VISC   = ((1.0-X) * VISCGP + X * VISCFP)/VISCFP/VISCGP;
/* SPEED OF SOUND */
	RATIO = DHFDTP(P,TS)/TK;
        SPEEDFP=sqrt( ABS(RATIO/(DRFDPT(P,TS)*1.E-06*RATIO -
                            (VFP*DRFDTP(P,TS))*(VFP*DRFDTP(P,TS))*1.E-03)));
	RATIO = DHGDTP(P,TS)/TK;
	SPEEDGP=sqrt( ABS(RATIO/(DRGDPT(P,TS)*1.E-06*RATIO -
                            (VGP*DRGDTP(P,TS))*(VGP*DRGDTP(P,TS))*1.E-03)));
/* Speed of sound of the mixture: 1/SPEED^2 = 1/SPEEDFP^2 +1/SPEEDGP^2 */
      SPEED=(X*SPEEDFP*SPEEDFP+(1.-X)*SPEEDGP*SPEEDGP)/
            (SPEEDFP*SPEEDFP*SPEEDGP*SPEEDGP);        /* new !!! */
/* THERMAL CONDUCTIVITY */   /* new !!! */
      TCFP = TCF(P);
      TCGP = TCG(P);
      TC   = (1.0-X)*TCFP + X*TCGP;
/* SURFACE TENSION  */
      SIGMA=SIGMAF(TS);    /* new !!! */
} /* end of function SAT */

/************************************************************************
* SUBCOOLED PROPERTY ROUTINE                                                                               
************************************************************************/
      
void SUB(float P,float T)
{
    /* function prototypes */
    
    float VFSUB (float P,float T);
    float HFSUB (float P,float T);
    float SFSUB (float P,float T);
    float CPFSUB(float P,float T);
    float DRFDTP(float P,float T);
    float DHFDTP(float P,float T);
    float DSFDTP(float P,float T);
    float DRFDPT(float P,float T);
    float DHFDPT(float P,float T);
    float DSFDPT(float P,float T);
    float DTDP(float P);
    float DPDT(float T);
    float VIFSUB(float P,float T);
    float TCFSUB(float P,float T);
    float SIGMAF(float T);
    float TK,RATIO;
/* SUBCOOLED, THEREFORE NO VOID OR QUALITY */
/* CALCULATE PHASE PROPERTIES */
    X=0.0;
    VOID=0.0;
    VFP=VFSUB(P,T);
    VGP=0.0;
    HFP=HFSUB(P,T);
    HGP=0.0;
    SFP=SFSUB(P,T);
    SGP=0.0;
    EFP=HFP-P*VFP*1000.0;
    EGP=0.0;
    RHOF=1.0/VFP;
    RHOG=0.0;
/* CALCULATE OVERALL SPECIFIC QUANTITIES */
    RHO=RHOF;
    SPECVOL=VFP;
    SPECH=HFP;
    SPECE=EFP;
    SPECS=SFP;
/* CALCULATE TOTAL QUANTITIES */
    TOTM=RHO*VOL;
    TOTH=SPECH*TOTM;
    TOTE=SPECE*TOTM;
    TOTS=SPECS*TOTM;
/* CALCULATE SLOPES */
    SRFDT = DRFDTP(P,T);
    SHFDT = DHFDTP(P,T);
    SSFDT = DSFDTP(P,T);
    SRFDP = DRFDPT(P,T);
    SHFDP = DHFDPT(P,T);
    SSFDP = DSFDPT(P,T);
    SRGDT = 0.0;
    SHGDT = 0.0;
    SRGDP = 0.0;
    SHGDP = 0.0;
    SSGDT = 0.0;
    SSGDP = 0.0;
    SPDT  = DPDT(T);
    STDP  = DTDP(P);
    SHDP  = SHFDP;
    SHDT  = SHFDT;
    SRDP  = SRFDP;
    SRDT  = SRFDT;
    SSDP  = SSFDP;
    SSDT  = SSFDT;
/* CALCULATE F FUNCTIONS */
    F1= RHO * SHFDT + SPECH * SRFDT;
    F2= - SRFDT;
    F3= - RHO*RHO * SHFDT;
    F4= 0.0;
    F5= SRFDP * SHFDT - SRFDT * SHFDP;
/* CALCULATE G FUNCTIONS */
    G1P= SHFDT / F5;
    G2P= -SRFDT / F5;
    G1T= -SHFDP / F5;
    G2T= SRFDP / F5;
/* BETA */
    BETAFP = -SRFDT/RHOF;    /* beta=(1/v)*(dv/dT)|p in 1/K */
    BETAGP = 0.0;
    BETA = BETAFP;
/* Ak */
        
    AKFP = SRFDP/RHOF;       /* alfa =-(1/v)*(dv/dp)|T in 1/MPa */
    AKGP = 0.0;
    AK = AKFP;
/* Cp */
    CPFP = CPFSUB(P,T);
    CPGP = 0.0;
    CP = CPFP;
/* Cv */
    TK = T+273.16;
    CVFP = CPFP - BETAFP*BETAFP * TK * VFP / AKFP *1.E3;
    CVGP = 0.0;
    CV = CVFP;
/* DYNAMIC VISCOSITY */
    VISCFP = VIFSUB(P,T);
    VISCGP = 0.0;
    VISC   = VISCFP;
/* SPEED OF SOUND */
/* RATIO = CPFP/TK;*/
                
    RATIO = SHFDT/TK;
    SPEEDFP=sqrt(ABS(RATIO/(SRFDP*1.E-06*RATIO-(VFP*SRFDT)*(VFP*SRFDT)*1.E-03)));
    SPEEDGP = 0.0;
    SPEED = SPEEDFP;
/* THERMAL CONDUCTIVITY  */
    TCFP = TCFSUB(P,T);
    TCGP = 0.0;
    TC   = TCFP;
/* SURFACE TENSION  */
    SIGMA=SIGMAF(T);
}  /* end of function SUB */

/************************************************************************
* SUPERHEATED PROPERTY ROUTINE                                                                             
************************************************************************/
          
void SUP(float P, float T)
{
        /* function prototypes */
        
    float VGSUP(float P,float T);
    float HGSUP(float P,float T);
    float SGSUP(float P,float T);
    float CPGSUP(float P,float T);
    float DRGDTP(float P,float T);
    float DHGDTP(float P,float T);
    float DSGDTP(float P,float T);
    float DRGDPT(float P,float T);
    float DHGDPT(float P,float T);
    float DSGDPT(float P,float T);
    float DTDP(float P);
    float DPDT(float P);
    float VIGSUP(float P,float T);
    float TCGSUP(float P,float T);
    float TK,RATIO;
/* SUPERHEATED, THEREFORE ALL VAPOUR */
/* CALCULATE PHASE PROPERTIES */
		X=1.0;
		VOID=1.0;
		VFP=0.0;
		VGP=VGSUP(P,T);
		HFP=0.0;
		HGP=HGSUP(P,T);
		SFP=0.0;
		SGP=SGSUP(P,T);
		EFP=0.0;
		EGP=HGP-P*VGP*1000.0;
		RHOF=0.0;
		RHOG=1.0/VGP;
/* CALCULATE OVERALL SPECIFIC QUANTITIES */
		RHO=RHOG;
                SPECVOL=VGP;
                SPECH=HGP;
                SPECE=EGP;
                SPECS=SGP;
/* CALCULATE TOTAL QUANTITIES */
                TOTM=RHO*VOL;
                TOTH=SPECH*TOTM;
                TOTE=SPECE*TOTM;
                TOTS=SPECS*TOTM;
/* CALCULATE SLOPES */
                SRGDT = DRGDTP(P,T);
                SHGDT = DHGDTP(P,T);
                SSGDT = DSGDTP(P,T);
                SRGDP = DRGDPT(P,T);
                SHGDP = DHGDPT(P,T);
                SSGDP = DSGDPT(P,T);
                SRFDT = 0.0;
                SHFDT = 0.0;
                SSFDT = 0.0;
                SRFDP = 0.0;
                SHFDP = 0.0;
                SSFDP = 0.0;
                SPDT  = DPDT(T);
                STDP  = DTDP(P);
                SHDP  = SHGDP;
                SHDT  = SHGDT;
                SRDP  = SRGDP;
                SRDT  = SRGDT;
                SSDP  = SSGDP;
                SSDT  = SSGDT;
/* F FUNCTIONS */
                F1= RHO * SHGDT + SPECH * SRGDT;
                F2= - SRGDT;
                F3= - RHO*RHO*SHGDT;
                F4= SRGDP * SHGDT - SRGDT * SHGDP;
                F5= 0.0;
/* G FUNCTIONS */
                G1P= SHGDT / F4;
                G2P= -SRGDT / F4;
                G1T= -SHGDP / F4;
                G2T= SRGDP / F4;
/* BETA */
                BETAFP = 0.0;
                BETAGP = -SRGDT/RHOG;   /* beta=(1/v)*(dv/dT)|p in 1/K */
                BETA = BETAGP;
/* Ak */
                AKFP = 0.0;
                AKGP = SRGDP/RHOG;      /* alfa =-(1/v)*(dv/dp)|T in 1/MPa */
                AK = AKGP;
/* Cp */
        
                CPFP = 0.0;
                CPGP = CPGSUP(P,T);
                CP = CPGP;
/* Cv */
                TK = T+273.16;
                CVFP = 0.0;
                CVGP = CPGP - BETAGP*BETAGP * TK * VGP / AKGP *1.E3;
                CV = CVGP;
/* DYNAMIC VISCOSITY */
                VISCFP = 0.0;
                VISCGP = VIGSUP(P,T);
                VISC   = VISCGP;
/* SPEED OF SOUND */
                
                /* RATIO = CPGP/TK;*/
                
                RATIO = SHGDT/TK;
                SPEEDFP = 0.0;
                SPEEDGP=sqrt(ABS(RATIO/(SRGDP*1.E-06*RATIO -
                                         (VGP*SRGDT)*(VGP*SRGDT)*1.E-03)));
                SPEED = SPEEDGP;
/* THERMAL CONDUCTIVITY  */
                TCGP = TCGSUP(P,T);
                TCFP = 0.0;
                TC   = TCGP;
/* SURFACE TENSION  */
                SIGMA=0.0;
}  /* end of function SUP */
/************************************************************************
* SUBCOOLED LIQUID - G FUNCTIONS                                                                           *
************************************************************************/
        
void GSUB(float P,float T,float *GP1,float *GP2,float *GT1,float *GT2)
{
    extern char *ERRH2O;
    /* function prototypes */
    
    float DRFDTP(float P,float T);
    float DHFDTP(float P,float T);
    float DRFDPT(float P,float T);
    float DHFDPT(float P,float T);

/* SLOPES */
 
        SRFDT = DRFDTP(P,T);
        SHFDT = DHFDTP(P,T);
        SRFDP = DRFDPT(P,T);
        SHFDP = DHFDPT(P,T);
        F5 = SRFDP * SHFDT - SRFDT * SHFDP;
/* GP FUNCTIONS */
        *GP1= SHFDT / F5;
        *GP2= -SRFDT / F5;
        F5= -SRFDP * SHFDT + SRFDT * SHFDP;
/* GT FUNCTIONS */
        *GT1= SHFDP / F5;
        *GT2= -SRFDP / F5;
}   /* end of function GSUB */
/************************************************************************
SUPERHEATED VAPOR  - G FUNCTIONS                                                                           *
************************************************************************/
void GSUP(float P,float T,float *GP1,float *GP2,float *GT1,float *GT2)
{
    extern char *ERRH2O;
    float DRGDTP(float P,float T);
    float DHGDTP(float P,float T);
    float DRGDPT(float P,float T);
    float DHGDPT(float P,float T);
/* SLOPES */
        SRGDT = DRGDTP(P,T);
        SHGDT = DHGDTP(P,T);
        SRGDP = DRGDPT(P,T);
        SHGDP = DHGDPT(P,T);
        F4= SRGDP * SHGDT - SRGDT * SHGDP;
/* GP FUNCTIONS */
        *GP1= SHGDT / F4;
        *GP2= -SRGDT / F4;
        F4= -SRGDP * SHGDT + SRGDT * SHGDP;
/* GT FUNCTIONS */
        *GT1= SHGDP / F4;
        *GT2= -SRGDP / F4;
}    /* end of function GSUP */
/************************************************************************
* Clip routines for pressure, temperature, enthalpy and density                   *
************************************************************************/
void CLIPP(float *P)
{
       float PMIN,PMAX;
        PMIN=.002;
        PMAX=21.55;
        *P=MIN(*P,PMAX);
        *P=MAX(*P,PMIN);
}     /* end of function CLIPP */

void CLIPT(float *T)
{
        float TMIN,TMAX;
        TMIN=17.5;
	TMAX=450.0;
        *T=MIN(*T,TMAX);
        *T=MAX(*T,TMIN);
}    /* end of function CLIPT */

void CLIPH(float *H)
{
        float HMIN,HMAX;
	HMIN=0.0;
	HMAX=3500.0;
        *H=MIN(*H,HMAX);
        *H=MAX(*H,HMIN);
}   /* end of function CLIPH */

void CLIPR(float *R)
{
        float RMIN,RMAX;
	RMIN=0.0001;
	RMAX=1100.0;
        *R=MIN(*R,RMAX);
        *R=MAX(*R,RMIN);
}    /* end of function CLIPR */

